// Windows stuff
#include "StdAfx.h"

// Defines / Structures / Variable stuff
#include "general.h"

// Variable declarations
#include "D3DFunc.h"

// DLL entry point
BOOL APIENTRY DllMain( HANDLE hModule, DWORD  ul_reason_for_call, LPVOID lpReserved ) { return TRUE; }

// Tells the DBP compiler to always include the Setup DLL. (Required for the plugin to function)
// NOTE: Not added to export table, undecorated form required.
// Doesn't look like we'll be needing this anymore
// UPDATE: Looks like we may need this after all
EXPORT char const * GetDependencyID(int iID)
{
	if (iID = 1)
		return "DBProSetupDebug.dll";
	else
		return NULL;
}

// Tell the DBP compiler how many DLLs we require.
// NOTE: Not added to export table, undecorated form required.
// Or this...
// UPDATE: And this
EXPORT int GetNumDependencies( void ) { return 1; }

// NOTE: Not added to export table, undecorated form required.
EXPORT void ReceiveCoreDataPtr( void *globPtr )
{
	g_pGlob = (GlobStruct *) globPtr;

	HWND hVWnd;
	hVWnd=GetDesktopWindow();
	GetWindowRect(hVWnd,&g_DesktopSize);
	g_DesktopDepth=g_pGlob->iScreenDepth;

	InitD3D();
}

// Grab functions from DBPs DLLs
void SetupFunctionPointers ( void )
{
	if (g_pGlob->g_Basic3Dmade)
	{
		CreateNewObject					= ( BISI ) GetProcAddress(g_pGlob->g_Basic3D, "?CreateNewObject@@YA_NHPADH@Z");
		GetObjectData					= ( OI ) GetProcAddress(g_pGlob->g_Basic3D, "?GetObjectA@@YAPAUsObject@@H@Z");
		SetupFVFData					= ( BODDD )GetProcAddress(g_pGlob->g_Basic3D, "?SetupMeshFVFData@@YA_NPAUsMesh@@KKK@Z");
		CalculateMeshBounds				= ( BO ) GetProcAddress(g_pGlob->g_Basic3D, "?CalculateMeshBounds@@YA_NPAUsMesh@@@Z");
		SetNewObjectFinalProperties		= ( VIF ) GetProcAddress(g_pGlob->g_Basic3D, "?SetNewObjectFinalProperties@@YA_NHM@Z");
		SetTexture						= ( VII ) GetProcAddress(g_pGlob->g_Basic3D, "?SetTexture@@YAXHH@Z");
		SetDetailTexture				= ( VII ) GetProcAddress(g_pGlob->g_Basic3D, "?SetDetailMap@@YAXHH@Z");
		GetExistEx						= ( II ) GetProcAddress(g_pGlob->g_Camera3D, "?GetExistEx@@YAHH@Z");
		//SetDetailTexture				= ( VII ) GetProcAddress(g_pGlob->g_Basic3D, "?SetDetailMap@@YAXHH@Z");
	}

	if (g_pGlob->g_Camera3Dmade)
	{
		GetInternalData					= ( DD ) GetProcAddress(g_pGlob->g_Camera3D, "?GetInternalData@@YAPAXH@Z");
		SetFOV							= ( VIF ) GetProcAddress(g_pGlob->g_Camera3D, "?SetFOV@@YAXHM@Z");
	}

}

// Initialize the plugin
void InitD3D()
{
	// Grab path to setup DLL
	char cPath[_MAX_PATH];
	strcpy(cPath, g_pGlob->pEXEUnpackDirectory);
	strcat(cPath, "\\DBProSetupDebug.dll");

	// Grab device pointer
	HMODULE hLib = LoadLibrary(cPath);
	OV getD3DDevice = (OV) GetProcAddress(hLib, "?GetDirect3DDevice@@YAPAUIDirect3DDevice9@@XZ");
	g_pD3DDevice = getD3DDevice();
	FreeLibrary(hLib);

	// Grab device pointer the easy way...
	// Which now seems to be broken with U6B4
	// g_pGlob->pCurrentBitmapSurface->GetDevice(&g_pD3DDevice);

	// Vertex buffer for drawing alpha 2D boxes
	if (g_pVB_Box)
		SAFE_RELEASE(g_pVB_Box);

	g_pD3DDevice->CreateVertexBuffer(4 * sizeof(TransformedVertex), D3DUSAGE_DYNAMIC | D3DUSAGE_WRITEONLY, D3DFVF_XYZRHW | D3DFVF_DIFFUSE, D3DPOOL_DEFAULT, &g_pVB_Box, NULL);

	// Create sprite for 2D text
	if (g_pTextSprite != NULL)
		SAFE_RELEASE(g_pTextSprite);
	HRESULT hr = D3DXCreateSprite(g_pD3DDevice, &g_pTextSprite);

	// Default font
	SetFont(1, "Arial", 15, 0, 0, 0);
	for (int i = 1; i < 20; i++)
		fiFont[i] = fiFont[0];

	// Grab Pro function pointers
	SetupFunctionPointers();

	// Default drawing colour
	g_Forecolour = D3DCOLOR_ARGB(255, 255, 255, 255);
}

// Set global drawing colour
void DrawColor(int nRed, int nGreen, int nBlue, int nAlpha)
{
	g_Forecolour = D3DCOLOR_ARGB(nAlpha, nRed, nGreen, nBlue);
}

// Return colour with alpha
DWORD D3DRGBA(int iR, int iG, int iB, int iA)
{
	return D3DCOLOR_ARGB(iA, iR, iG, iB);
}

// Turn a DX Mesh into a Pro object
// TODO: Go over this
void MakeObjectFromDXMesh( int iID,ID3DXMesh *ppMesh )
{
    HRESULT  hr;
    int   iFinalMeshCount   = 1;
    int   iMesh     = 0;
    sObject* pObject     = NULL;
    sMesh*  pMesh     = NULL;  // temporary mesh pointer
    sFrame*  pFrame     = NULL;  // temporary frame pointer
    sMesh**  ppMeshListFromObject = NULL;  // mesh list array
    DWORD  dwIndexCount   = 0;  // index count
    DWORD  dwVertexCount   = 0;  // vertex count
    BYTE*  pVertices    =NULL;
    DWORD  nVertSize;
    LPDIRECT3DINDEXBUFFER9 pMeshIndexBuffer;
    short *pIndices;

    CreateNewObject( iID, "3DText", iFinalMeshCount );

    pObject = GetObjectData ( iID );

    if ( !pObject )
    {
        // error
    }

    // now get a pointer to the first mesh and first frame, finally
    // create a temporary mesh list so we can access by array and not
    // through the child structure, we can't use the object mesh list
    // at this stage because it hasn't been created yet

    pMesh     = pObject->pFrame->pMesh;
    pFrame        = pObject->pFrame;
    ppMeshListFromObject = new sMesh* [ iFinalMeshCount ];

    if ( !ppMeshListFromObject )
    {
        // error
    }

    // store pointers to meshes in the mesh list

    for ( iMesh = 0; iMesh < iFinalMeshCount; iMesh++ )
    {
        // store mesh pointer

        ppMeshListFromObject [ iMesh ] = pMesh;

        // move temporary pointers to child frames

        if ( pFrame->pChild )
        {
            pFrame = pFrame->pChild;
            pMesh  = pFrame->pMesh;
        }
    }

    // go through all meshes in list

    for ( iMesh = 0; iMesh < iFinalMeshCount; iMesh++ )
    {
        // get index and vertex count of the block

        dwIndexCount=ppMesh->GetNumFaces() * 3;
        dwVertexCount = ppMesh->GetNumVertices();

        // set up the new mesh

        SetupFVFData ( ppMeshListFromObject [ iMesh ], D3DFVF_DBPMESH, dwVertexCount, dwIndexCount );

        sVertex* pVertex = ( sVertex* ) ppMeshListFromObject [ iMesh ]->pVertexData;

        hr=ppMesh->LockVertexBuffer(D3DLOCK_READONLY, (LPVOID*)&pVertices);
        nVertSize=D3DXGetFVFVertexSize(ppMesh->GetFVF());

        //Get pointer to Index buffer and lock the index buffer in prder to access it

        ppMesh->GetIndexBuffer(&pMeshIndexBuffer);
        pMeshIndexBuffer->Lock(0, 3 * ppMesh->GetNumFaces() * sizeof(short), (void **)&pIndices, D3DLOCK_READONLY);

        // copy vertices from block into our new mesh

        for ( int iVertex = 0; iVertex < (int)dwVertexCount; iVertex++ )
        {
            sVertex p = *(sVertex*)(pVertices + iVertex*nVertSize);
            pVertex [ iVertex ].x         = p.x;
            pVertex [ iVertex ].y         = p.y;
            pVertex [ iVertex ].z         = p.z;

            pVertex [ iVertex ].nx         = p.nx;
            pVertex [ iVertex ].ny         = p.ny;
            pVertex [ iVertex ].nz         = p.nz;

            pVertex [ iVertex ].tu        = p.tu;
            pVertex [ iVertex ].tv        = p.tv;
        }
        ppMesh->UnlockVertexBuffer();

        // copy indices from block into mesh

        for ( int iIndex = 0; iIndex < (int)dwIndexCount; iIndex++ )
            ppMeshListFromObject [ iMesh ]->pIndices [ iIndex ] = pIndices[iIndex];

        pMeshIndexBuffer->Unlock();

        // set up mesh properties, no lights as we use diffuse, triangle list for primitive type, store
        // the vertex count and finally the primitive count

        ppMeshListFromObject [ iMesh ]->bLight           = true;
        ppMeshListFromObject [ iMesh ]->iPrimitiveType   = D3DPT_TRIANGLELIST;
        ppMeshListFromObject [ iMesh ]->iDrawVertexCount = ppMeshListFromObject [ iMesh ]->dwVertexCount;
        ppMeshListFromObject [ iMesh ]->iDrawPrimitives  = ppMesh->GetNumFaces ( );
        ppMeshListFromObject [ iMesh ]->iMeshType   = 1;

        // calculate the mesh bounds

        CalculateMeshBounds ( ppMeshListFromObject [ iMesh ] );
    }

    // set the final properties of the object and apply a blank texture

    SetNewObjectFinalProperties ( iID, 0.0f );
    SetTexture                  ( iID, 0 );

    // safely delete the temporary mesh list

    SAFE_DELETE_ARRAY ( ppMeshListFromObject );
    SAFE_RELEASE ( ppMesh );
}

void GetCameraNormal ( int iCamera )
{
		DWORD dwCameraInternalData = GetInternalData((DWORD)iCamera);
	if (!dwCameraInternalData)
	{
		// Invalid camera
		*(DWORD*)g_pGlob->g_pErrorHandlerRef = (DWORD)7203;
		return;
	}

	g_pViewPort.X = *(int*)(dwCameraInternalData+200);
    g_pViewPort.Y = *(int*)(dwCameraInternalData+204);
    g_pViewPort.Width = *(int*)(dwCameraInternalData+208);
    g_pViewPort.Height = *(int*)(dwCameraInternalData+212);
	// TODO: Should these be CameraNear/Far?
    g_pViewPort.MinZ = 0.0f;
    g_pViewPort.MaxZ = 1.0f;

	D3DXMATRIX matWorld;
	D3DXMatrixIdentity(&matWorld);

	D3DXMATRIX matProj = *(D3DXMATRIX*)(dwCameraInternalData);
    D3DXMATRIX matView = *(D3DXMATRIX*)(dwCameraInternalData+64);

	D3DXVECTOR3 v1;

	v1.x=(float)(g_pViewPort.X + g_pViewPort.Width / 2);
	v1.y=(float)(g_pViewPort.Y + g_pViewPort.Height / 2);
	v1.z=1.0f;

	D3DXVec3Unproject( &g_RayDir, &v1, &g_pViewPort, &matProj, &matView, &matWorld );
	D3DXVec3Normalize( &g_RayDir, &g_RayDir );
}
void GetCameraPick( int iCamera, int x, int y, float iDist )
{
		DWORD dwCameraInternalData = GetInternalData((DWORD)iCamera);
	if (!dwCameraInternalData)
	{
		// Invalid camera
		*(DWORD*)g_pGlob->g_pErrorHandlerRef = (DWORD)7203;
		return;
	}

	float CameraNear = *(float*)(dwCameraInternalData+344);
    float CameraFar = *(float*)(dwCameraInternalData+348);

	g_pViewPort.X = *(int*)(dwCameraInternalData+200);
    g_pViewPort.Y = *(int*)(dwCameraInternalData+204);
    g_pViewPort.Width = *(int*)(dwCameraInternalData+208);
    g_pViewPort.Height = *(int*)(dwCameraInternalData+212);
	// TODO: Should these be CameraNear/Far?
    g_pViewPort.MinZ = 0.0f;
    g_pViewPort.MaxZ = 1.0f;

	D3DXMATRIX matWorld;
	D3DXMatrixIdentity(&matWorld);

	D3DXMATRIX matProj = *(D3DXMATRIX*)(dwCameraInternalData);
    D3DXMATRIX matView = *(D3DXMATRIX*)(dwCameraInternalData+64);

	//float sngA = matProj._33;
	//float sngB = matProj._43;
	//float sngPlaneNear = -sngB / sngA;
	//float sngPlaneFar = sngB / (1 - sngA);

	D3DXVECTOR3 v1,v2;

	v1.x=(float)x;
	v1.y=(float)y;
	v1.z=0.0f;

	v2.x=(float)x;
	v2.y=(float)y;
	v2.z=1.0f;

	D3DXVec3Unproject( &g_RayOrigin, &v1, &g_pViewPort, &matProj, &matView, &matWorld );
	D3DXVec3Unproject( &g_RayDir, &v2, &g_pViewPort, &matProj, &matView, &matWorld );
	D3DXVec3Subtract( &g_RayDir, &g_RayDir, &g_RayOrigin );
	D3DXVec3Normalize( &g_RayDir, &g_RayDir );
	D3DXVec3Scale( &g_RayDir, &g_RayDir, iDist );
	D3DXVec3Add( &g_RayDir, &g_RayOrigin, &g_RayDir );
}

DWORD GetCameraX ( void )
{
//`	if (iValue)
		return *(DWORD*)&g_RayDir.x;
//	else
//		return *(DWORD*)&g_RayOrigin.x;

}
DWORD GetCameraY ( void )
{
//	if (iValue)
		return *(DWORD*)&g_RayDir.y;
//	else
//		return *(DWORD*)&g_RayOrigin.y;

}
DWORD GetCameraZ ( void )
{
//	if (iValue)
		return *(DWORD*)&g_RayDir.z;
//	else
//		return *(DWORD*)&g_RayOrigin.z;

}

int GetDesktopWidth ( void )
{
	return(g_DesktopSize.right);
}

int GetDesktopHeight ( void )
{
	return(g_DesktopSize.bottom);
}

int GetDesktopDepth ( void )
{
	return(g_DesktopDepth);
}

void SetOrthoProjection ( int iCamera, float fWidth )
{
		DWORD dwCameraInternalData = GetInternalData((DWORD)iCamera);
	if (!dwCameraInternalData)
	{
		// Invalid camera
		*(DWORD*)g_pGlob->g_pErrorHandlerRef = (DWORD)7203;
		return;
	}

	float CameraNear = *(float*)(dwCameraInternalData+344);
    float CameraFar = *(float*)(dwCameraInternalData+348);

	g_pViewPort.X = *(int*)(dwCameraInternalData+200);
    g_pViewPort.Y = *(int*)(dwCameraInternalData+204);
    g_pViewPort.Width = *(int*)(dwCameraInternalData+208);
    g_pViewPort.Height = *(int*)(dwCameraInternalData+212);

	//Save the camera details
	g_SaveCameraDets.FOV=*(float*)(dwCameraInternalData+340);
	g_SaveCameraDets.CameraNear = *(float*)(dwCameraInternalData+344);
    g_SaveCameraDets.CameraFar = *(float*)(dwCameraInternalData+348);
	g_SaveCameraDets.pViewPort.X = *(int*)(dwCameraInternalData+200);
    g_SaveCameraDets.pViewPort.Y = *(int*)(dwCameraInternalData+204);
    g_SaveCameraDets.pViewPort.Width = *(int*)(dwCameraInternalData+208);
    g_SaveCameraDets.pViewPort.Height = *(int*)(dwCameraInternalData+212);
	// TODO: Should these be CameraNear/Far?
    //g_pViewPort.MinZ = 0.0f;
    //g_pViewPort.MaxZ = 1.0f;

	D3DXMATRIX matProj = *(D3DXMATRIX*)(dwCameraInternalData);

	//float fWidth=g_pViewPort.Width/fZoom;
	float fHeight=fWidth/((float)g_pViewPort.Width/(float)g_pViewPort.Height);

	D3DXMatrixOrthoLH(&matProj, fWidth, fHeight, 0, CameraFar);
	*(D3DXMATRIX*)(dwCameraInternalData)=matProj;
}

void RestoreCamera ( int iCamera )
{
		DWORD dwCameraInternalData = GetInternalData((DWORD)iCamera);
	if (!dwCameraInternalData)
	{
		// Invalid camera
		*(DWORD*)g_pGlob->g_pErrorHandlerRef = (DWORD)7203;
		return;
	}

	float CameraFOV = *(float*)(dwCameraInternalData+340);

	SetFOV(iCamera, CameraFOV);
}

int CursorInCamera ( int iFirstCam, int iLastCam )
{
	int iCamera;
	DWORD nMouseX = (int)g_pGlob->iWindowsMouseX;
	DWORD nMouseY = (int)g_pGlob->iWindowsMouseY;

	for (iCamera=iFirstCam; iCamera<iLastCam+1 ; iCamera++)
	{
		if (GetExistEx(iCamera)==1)
		{
			DWORD dwCameraInternalData = GetInternalData((DWORD)iCamera);
			D3DVIEWPORT9	pViewPort;
			pViewPort.X = *(int*)(dwCameraInternalData+200);
			pViewPort.Y = *(int*)(dwCameraInternalData+204);
			pViewPort.Width = *(int*)(dwCameraInternalData+208);
			pViewPort.Height = *(int*)(dwCameraInternalData+212);
			if (nMouseX>pViewPort.X && nMouseX<pViewPort.X + pViewPort.Width && nMouseY>pViewPort.Y && nMouseY<pViewPort.Y + pViewPort.Height)
			{
				return(iCamera);
				break;
			}
		}
	}
	return(-1);
}

inline int RoundValue(float param)
{
    // Uses the FloatToInt functionality
    int a;
    int *int_pointer = &a;

    __asm  fld  param
    __asm  mov  edx,int_pointer
    __asm  FRNDINT
    __asm  fistp dword ptr [edx];

    return a;
}
DWORD Round(float number, int num_digits)
{
    float doComplete5i, doComplete5(number * powf(10.0f, (float) (num_digits + 1)));
    
    if(number < 0.0f)
        doComplete5 -= 5.0f;
    else
        doComplete5 += 5.0f;
    
    doComplete5 /= 10.0f;
    modff(doComplete5, &doComplete5i);
    
    doComplete5i=doComplete5i / powf(10.0f, (float) num_digits);
	return *(DWORD*)&doComplete5i;
}
